/* $Workfile: OsaLinuxAsp.c $   $Revision: 12 $    $Modtime: 5/13/09 1:49p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 2007, All Rights Reserved

*******************************************************************************/

/******************************************************************************
  
  Driver exported entry/exit functions.

  Some of the Linux kernel headers don't like to be compiled with g++ (C++)
  so we need a file that is pure C. This file has two purposes: it implements
  the two Linux module specific entry exit routines, init_module and
  cleanup_module, the file operation callbacks, and also some routines that's
  needed for proc operations.

******************************************************************************/
// This is the only place where this file should be included
// It's unlikely it will be needed anywhere else but if so
// __NO_VERSION__ must be defined before inclusion at the
// other places

// Only include with DaVinci
#ifdef DAVINCI


#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/dma-mapping.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>


// DBG is not defined for C files only CPP
#if OSA_DBG
#define DBG 1
#endif
#include "OsaLinuxRtl.h"  // For OSA_DPRINTF
#if OSA_DBG
#undef DBG
#endif

#include "OsaLinuxAsp.h"
#include "OsaLinuxDmp.h"


#include <asm/arch/hardware.h>
#include <asm/arch/edma.h>
#include <asm/arch/mcbsp.h>

#define OSA_ASP_INVALID_LCH -1

#define DAVINCI_DMA_REGISTER_BASE DAVINCI_DMA_3PCC_BASE

#define CHIP_MEM  0x11104000  // Start of VICP memory (32k long?)

//#define OSA_ASP_USE_DDR_MEM_FOR_BUFFERS  // Undef to use VICP memory instead of DDR2 memory

#define DAVINCI_MSTPRI0 0x01C4003C // Bus master priority control 0. For details, see Section 3.5.1, Switched Central Resource (SCR) Bus Priorities.
#define DAVINCI_MSTPRI1 0x01C40040 // Bus master priority control 1. For details, see Section 3.5.1, Switched Central Resource (SCR) Bus Priorities.

#define DAVINCI_FIQ0     0x01C48000    // Interrupt Status 0 [Interrupt Status of INT[31:0] (If Mapped to FIQ)]
#define DAVINCI_FIQ1     0x01C48004    // Interrupt Status 1 [Interrupt Status of INT[63:32] (If Mapped to FIQ)]
#define DAVINCI_IRQ0     0x01C48008    // Interrupt Status 0 [Interrupt Status of INT[31:0] (If Mapped to IRQ)]
#define DAVINCI_IRQ1     0x01C4800C    // Interrupt Status 1 [Interrupt Status of INT[63:32] (If Mapped to IRQ)]
#define DAVINCI_FIQENTRY 0x01C48010    // Entry Address [28:0] for Valid FIQ Interrupt
#define DAVINCI_IRQENTRY 0x01C48014    // Entry Address [28:0] for Valid IRQ Interrupt
#define DAVINCI_EINT0    0x01C48018    // Interrupt Enable Register 0
#define DAVINCI_EINT1    0x01C4801C    // Interrupt Enable Register 1
#define DAVINCI_INCTL    0x01C48020    // Interrupt Operation Control Register
#define DAVINCI_EABASE   0x01C48024    // Interrupt Entry Table Base Address Register
#define DAVINCI_INTPRI0  0x01C48030    // Interrupt 0-7 Priority Select
#define DAVINCI_INTPRI1  0x01C48034    // Interrupt 8-15 Priority Select
#define DAVINCI_INTPRI2  0x01C48038    // Interrupt 16-23 Priority Select
#define DAVINCI_INTPRI3  0x01C4803C    // Interrupt 24-31 Priority Select. This register holds the interrupt priority for the ASP interrupts
#define DAVINCI_INTPRI4  0x01C48040    // Interrupt 32-39 Priority Select
#define DAVINCI_INTPRI5  0x01C48044    // Interrupt 40-47 Priority Select
#define DAVINCI_INTPRI6  0x01C48048    // Interrupt 48-55 Priority Select
#define DAVINCI_INTPRI7  0x01C4804C    // Interrupt 56-63 Priority Select

#define DAVINCI_AWCCR 0x01E00004  // Asynchronous Wait Cycle Configuration Register
#define DAVINCI_A1CR  0x01E00010  // Asynchronous 1 Configuration Register (CS2 Space)


#define MCBSP_DRR    0x01E02000   // Data receive
#define MCBSP_DXR    0x01E02004   // Data transmit
#define MCBSP_SPCR   0x01E02008   // Serial port control
#define MCBSP_RCR    0x01E0200C   // Receive control
#define MCBSP_XCR    0x01E02010   // Transmit control
#define MCBSP_SRGR   0x01E02014   // Sample rate generator
#define MCBSP_PCR    0x01E02024   // Pin control

/********************** MCBSP SPCR bit (32-bit) definitions ***********************/
#define MCBSP_SPCR_RRST			      0x00000001    // Bit 0
#define MCBSP_SPCR_RRDY			      0x00000002    // Bit 1
#define MCBSP_SPCR_RFULL		  	  0x00000004    // Bit 2
#define MCBSP_SPCR_RSYNC_ERR		  0x00000008    // Bit 3
#define MCBSP_SPCR_RINTM(value)		((value)<<4)	/* bits 4:5 */
#define MCBSP_SPCR_ABIS			      0x00000040    // Bit 6
#define MCBSP_SPCR_DXENA			    0x00000080    // Bit 7
#define MCBSP_SPCR_CLKSTP(value)	((value)<<11)	/* bits 11:12 */
#define MCBSP_SPCR_RJUST(value)		((value)<<13)	/* bits 13:14 */
#define MCBSP_SPCR_DLB			      0x00008000    // Bit 16

#define MCBSP_SPCR_XRST		        0x00010000    // Bit 17
#define MCBSP_SPCR_XRDY		        0x00020000    // Bit 18
#define MCBSP_SPCR_XEMPTY		      0x00040000    // Bit 19
#define MCBSP_SPCR_XSYNC_ERR	    0x00080000    // Bit 20
#define MCBSP_SPCR_XINTM(value)	  ((value)<<20)	/* bits 20:21 */
#define MCBSP_SPCR_GRST		        0x00400000    // Bit 22
#define MCBSP_SPCR_FRST		        0x00800000    // Bit 23
#define MCBSP_SPCR_SOFT		        0x01000000    // Bit 24
#define MCBSP_SPCR_FREE		        0x02000000    // Bit 25



typedef enum {
  ASP_RX_DIR,   // Reading from DRR
  ASP_TX_DIR,   // Writing to DXR
} AspDirectionT;

extern void davinci_clean_channel(int ch_no); // For some reason this function was not declared in <asm/arch/edma.h>

static int OsaLinuxCreateAndConfigureDmaChannels(void) ;

static void OsaLinuxAspFreeDmaChannels(void);

static void OsaLinuxAspMcbspStart(int bStartRx, int bStartTx);

static void OsaLinuxAspFreeBuffers(void);

static int OsaLinuxAspCreateDmaChannel(
  AspDirectionT eDirection, // ASP_RX_DIR, or ASP_TX_DIR
  dma_addr_t nPhysBufAddr,  // Physical address of the buffer for this channel
  int nDevId,               // DMA channel number. DAVINCI_DMA_MCBSP1_RX for master or DAVINCI_EDMA_PARAM_ANY for slave channels
  char *sqName,             // Name of the dma channel in human readable format
  void (*pfCallback) (int, unsigned short, void *),  // Channel callback function (valid only if you are requesting for a DMA MasterChannel)
  int  nTccRequested,       // The TCC that we request to be used (Ignored for Master channels)
  int  *nTccActual,         // Actual TCC used for the PaRAM
  int  *nLch                // Logical Channel Id 0-127 (PaRAM set)
);

static inline u32 OsaLinuxAspReadReg32(u32 address);
static inline void OsaLinuxAspWriteReg32(u32 address, u32 value);

// Interrupt service functions
static void OsaLinuxAspIrqRxMaster(int lch, u16 ch_status, void *data);
static void OsaLinuxAspIrqRxPing(int lch, u16 ch_status, void *data);
static void OsaLinuxAspIrqRxPong(int lch, u16 ch_status, void *data);
static void OsaLinuxAspIrqTxMaster(int lch, u16 ch_status, void *data);
static void OsaLinuxAspIrqTxPing(int lch, u16 ch_status, void *data);
static void OsaLinuxAspIrqTxPong(int lch, u16 ch_status, void *data);


static int g_nLchMcbspRx         = OSA_ASP_INVALID_LCH; // Master channel Rx
static int g_nLchMcbspRxLinkPong = OSA_ASP_INVALID_LCH;
static int g_nLchMcbspRxLinkPing = OSA_ASP_INVALID_LCH;
static int g_nTccMcbspRx         = 0; 
static int g_nTccMcbspRxLinkPong = 0;
static int g_nTccMcbspRxLinkPing = 0;

static int g_nLchMcbspTx         = OSA_ASP_INVALID_LCH; // Master channel Tx
static int g_nLchMcbspTxLinkPong = OSA_ASP_INVALID_LCH;
static int g_nLchMcbspTxLinkPing = OSA_ASP_INVALID_LCH;
static int g_nTccMcbspTx         = 0; // Master channel Tx
static int g_nTccMcbspTxLinkPong = 0;
static int g_nTccMcbspTxLinkPing = 0;

static int g_bRxPingBufReady = 0;  // Boolean, If true then we ping buf is ready to be accessed. If false, the buffer is not ready to be accessed
static int g_bRxPongBufReady = 0;  // Boolean, If true then we pong buf is ready to be accessed. If false, the buffer is not ready to be accessed
static int g_bTxPingBufReady = 0;  // Boolean, If true then we ping buf is ready to be accessed. If false, the buffer is not ready to be accessed
static int g_bTxPongBufReady = 0;  // Boolean, If true then we pong buf is ready to be accessed. If false, the buffer is not ready to be accessed

static unsigned g_nCountRxMaster = 0;
static unsigned g_nCountTxMaster = 0;

static unsigned g_nCountRxPing   = 0;
static unsigned g_nCountRxPong   = 0;
static unsigned g_nCountTxPing   = 0;
static unsigned g_nCountTxPong   = 0;

static u32 g_nSavedAwccr = 0;

static int g_cbBufSize = 0;  // The size (in bytes) of each ping and pong buffer. This variable gets set in the OsaLinuxAspInit() function.

static void *g_pCallbackContext = 0;
static void (*g_pCallbackFuncTxIrq)(unsigned long, void *) = 0;
static void (*g_pCallbackFuncRxIrq)(unsigned long, void *) = 0;

typedef enum {
  OSA_ASP_STATE_UNINITIALIZED, 
  OSA_ASP_STATE_INITIALIZED, 
  OSA_ASP_STATE_OPENED, 
  OSA_ASP_STATE_ENABLED, 
} OsaAspStateT;

static OsaAspStateT g_eDeviceState = OSA_ASP_STATE_UNINITIALIZED;

volatile edmacc_regs *ptr_edmacc_regs = NULL;

#define OSA_ASP_N_BUFFERS 4

#define OSA_ASP_BUF_TX_PING 0
#define OSA_ASP_BUF_RX_PING 1
#define OSA_ASP_BUF_TX_PONG 2
#define OSA_ASP_BUF_RX_PONG 3

#define OSA_ASP_TCC_RX_LINK_PONG 11
#define OSA_ASP_TCC_RX_LINK_PING 12

#define OSA_ASP_TCC_TX_LINK_PONG 21
#define OSA_ASP_TCC_TX_LINK_PING 22

static struct {
  u8 *virt;
  dma_addr_t phys;
} g_sqDmaBuffers[OSA_ASP_N_BUFFERS];


#define ASP_MCBSP DAVINCI_MCBSP1


/*
** SPCR = 0x00000000  // In reset
** RCR  = 0x00a113a0
** XCR  = 0x00a013a0
** SRGR = 0x21010000
** PRC  = 0x00000080
*/

// 2 is high word bits (31-16)
// 1 is low word (bits 15-0)
static struct davinci_mcbsp_reg_cfg g_sMcBspInitialConfig = {
  // Default setting is for the Alvis-4-CSI
  // Frame Sync as a negative pulse output 488ns width, 125 us period assuming CLKX is 8.192 MHz (generated from the incoming CLKX pin)
	.spcr2 = XINTM(2),  // FRST and GRST in this register are set later by the OsaLinuxAspMcbspStart() function
	.spcr1 = RINTM(2),
	.rcr2  = RWDLEN2(DAVINCI_MCBSP_WORD_32) | RDATDLY(0),
	.rcr1  = RFRLEN1(((128/4)-1)) | RWDLEN1(DAVINCI_MCBSP_WORD_32),
	.xcr2  = XWDLEN2(DAVINCI_MCBSP_WORD_32) | XDATDLY(0),
	.xcr1  = XFRLEN1(((128/4)-1)) | XWDLEN1(DAVINCI_MCBSP_WORD_32),
	.srgr2 = CLKSM | FSGM| FPER(0x3FF),
	.srgr1 = FWID(0x0) | CLKGDV(0),
	.pcr0  = FSXM | FSRM | SCLKME | FSXP | FSRP | CLKXP,
};


#define OSA_LINUX_CHECK_FOR_RFULL(spcr)       ((spcr & MCBSP_SPCR_RFULL)     ? OSA_ASP_DMA_OVER_UNDER_RUN_HW : 0)
#define OSA_LINUX_CHECK_FOR_RSYNC_ERROR(spcr) ((spcr & MCBSP_SPCR_RSYNC_ERR) ? OSA_ASP_DMA_SYNC_ERROR : 0)
#define OSA_LINUX_CHECK_FOR_XEMPTY(spcr)      (!(spcr & MCBSP_SPCR_XEMPTY)   ? OSA_ASP_DMA_OVER_UNDER_RUN_HW : 0)
#define OSA_LINUX_CHECK_FOR_XSYNC_ERROR(spcr) ((spcr & MCBSP_SPCR_XSYNC_ERR) ? OSA_ASP_DMA_SYNC_ERROR : 0)

/******************************************************************************
******************************************************************************/
static inline u32 OsaLinuxAspDecodeChStatus(u16 ch_status)
{
  u32 nStatus;
  switch (ch_status) {
  case DMA_COMPLETE:
    nStatus = OSA_ASP_DMA_COMPLETE;
    break;
  case DMA_CC_ERROR:
  case DMA_TC1_ERROR:
  case DMA_TC2_ERROR:
    nStatus = OSA_ASP_DMA_ERROR;
    break;
  default:
    nStatus = 0;
    break;
  }
  return nStatus;
}  

/******************************************************************************
******************************************************************************/
static void OsaLinuxAspIrqRxMaster(int lch, u16 ch_status, void *data)
{
  g_nCountRxMaster++;
  OsaLinuxAspIrqRxPing(lch, ch_status, data);
}
/******************************************************************************
******************************************************************************/
static void OsaLinuxAspIrqRxPing(int lch, u16 ch_status, void *data)
{
  u32 nStatus;
  u32 spcr = OsaLinuxAspReadReg32(MCBSP_SPCR);

  g_nCountRxPing++;
  nStatus = OsaLinuxAspDecodeChStatus(ch_status);
  nStatus |= OSA_LINUX_CHECK_FOR_RSYNC_ERROR(spcr);
  nStatus |= OSA_LINUX_CHECK_FOR_RFULL(spcr);
  nStatus |= OSA_ASP_DMA_PING_BUF_READY;
  if (g_bRxPongBufReady) {
    nStatus |= OSA_ASP_DMA_OVER_UNDER_RUN_USER; // The user was not quick enough to read to the other buffer
    g_bRxPongBufReady = 0; // The DMA is now writing to this buffer, so it is not ready
  }
  g_bRxPingBufReady = 1;
  if (g_pCallbackFuncRxIrq) {
    (*g_pCallbackFuncRxIrq)(nStatus, g_pCallbackContext);
  }
}
/******************************************************************************
******************************************************************************/
static void OsaLinuxAspIrqRxPong(int lch, u16 ch_status, void *data)
{
  u32 nStatus;
  u32 spcr = OsaLinuxAspReadReg32(MCBSP_SPCR);

  g_nCountRxPong++;
  nStatus = OsaLinuxAspDecodeChStatus(ch_status);
  nStatus |= OSA_LINUX_CHECK_FOR_RSYNC_ERROR(spcr);
  nStatus |= OSA_LINUX_CHECK_FOR_RFULL(spcr);
  nStatus |= OSA_ASP_DMA_PONG_BUF_READY;
  if (g_bRxPingBufReady) {
    nStatus |= OSA_ASP_DMA_OVER_UNDER_RUN_USER; // The user was not quick enough to read to the other buffer
    g_bRxPingBufReady = 0; // The DMA is now writing to this buffer, so it is not ready
  }
  g_bRxPongBufReady = 1;
  if (g_pCallbackFuncRxIrq) {
    (*g_pCallbackFuncRxIrq)(nStatus, g_pCallbackContext);
  }
}
/******************************************************************************
******************************************************************************/
static void OsaLinuxAspIrqTxMaster(int lch, u16 ch_status, void *data)
{
  g_nCountTxMaster++;
  OsaLinuxAspIrqTxPing(lch, ch_status, data);
}
/******************************************************************************
******************************************************************************/
static void OsaLinuxAspIrqTxPing(int lch, u16 ch_status, void *data)
{
  u32 nStatus;
  u32 spcr = OsaLinuxAspReadReg32(MCBSP_SPCR);

  g_nCountTxPing++;
  nStatus = OsaLinuxAspDecodeChStatus(ch_status);
  nStatus |= OSA_LINUX_CHECK_FOR_XSYNC_ERROR(spcr);
  nStatus |= OSA_LINUX_CHECK_FOR_XEMPTY(spcr);
  nStatus |= OSA_ASP_DMA_PING_BUF_READY;
  if (g_bTxPongBufReady) {
    nStatus |= OSA_ASP_DMA_OVER_UNDER_RUN_USER; // The user was not quick enough to read to the other buffer
    g_bTxPongBufReady = 0; // The DMA is now writing to this buffer, so it is not ready
  }
  g_bTxPingBufReady = 1;
  if (g_pCallbackFuncTxIrq) {
    (*g_pCallbackFuncTxIrq)(nStatus, g_pCallbackContext);
  }
}
/******************************************************************************
******************************************************************************/
static void OsaLinuxAspIrqTxPong(int lch, u16 ch_status, void *data)
{
  u32 nStatus;
  u32 spcr = OsaLinuxAspReadReg32(MCBSP_SPCR);

  g_nCountTxPong++;
  nStatus = OsaLinuxAspDecodeChStatus(ch_status);
  nStatus |= OSA_LINUX_CHECK_FOR_XSYNC_ERROR(spcr);
  nStatus |= OSA_LINUX_CHECK_FOR_XEMPTY(spcr);
  nStatus |= OSA_ASP_DMA_PONG_BUF_READY;
  if (g_bTxPingBufReady) {
    nStatus |= OSA_ASP_DMA_OVER_UNDER_RUN_USER; // The user was not quick enough to read to the other buffer
    g_bTxPingBufReady = 0; // The DMA is now writing to this buffer, so it is not ready
  }
  g_bTxPongBufReady = 1;
  if (g_pCallbackFuncTxIrq) {
    (*g_pCallbackFuncTxIrq)(nStatus, g_pCallbackContext);
  }
}


   

/******************************************************************************
  OsaLinuxAspAllocBuffers
    return 1 on success
******************************************************************************/
int OsaLinuxAspAllocBuffers(int cbBufSize)
{
  int i;

  if (g_eDeviceState != OSA_ASP_STATE_OPENED) {
    OSA_EPRINTF(("OsaLinuxAspChangeBufSize: The buffer size can not be changed in this state.\n"));
    return -EINVAL;
  }
  g_cbBufSize = cbBufSize;

  OsaLinuxAspFreeBuffers();

  for(i = 0; i < OSA_ASP_N_BUFFERS; i++) {
#ifdef OSA_ASP_USE_DDR_MEM_FOR_BUFFERS
    g_sqDmaBuffers[i].virt = (u8 *)dma_alloc_coherent(NULL, 
                                                g_cbBufSize, 
                                                &g_sqDmaBuffers[i].phys, 
                                                GFP_KERNEL);
#else // Use VICP memory
    g_sqDmaBuffers[i].phys = CHIP_MEM + g_cbBufSize*i;
    g_sqDmaBuffers[i].virt = ioremap(g_sqDmaBuffers[i].phys, g_cbBufSize);
#endif
    if(g_sqDmaBuffers[i].virt == NULL) {
      OSA_EPRINTF(("OsaLinuxAspAllocBuffers() failed\n"));
      return 0;
    }

    OSA_DPRINTF(("g_sqDmaBuffers[%d].virt = 0x%p  phys = 0x%.8X \n", i, g_sqDmaBuffers[i].virt, g_sqDmaBuffers[i].phys));

#if 0
    memset(g_sqDmaBuffers[i].virt, 0, g_cbBufSize);
#else
    memset(g_sqDmaBuffers[i].virt, 0xA2, g_cbBufSize);
#endif
  }

  return 1;
}


/******************************************************************************
  OsaLinuxAspCreateDmaChannel
    return 1-.OK, 0->error
******************************************************************************/
static int OsaLinuxAspCreateDmaChannel(
  AspDirectionT eDirection, // ASP_RX_DIR, or ASP_TX_DIR
  dma_addr_t nPhysBufAddr,  // Physical address of the buffer for this channel
  int nDevId,               // DMA channel number. DAVINCI_DMA_MCBSP1_RX for master or DAVINCI_EDMA_PARAM_ANY for slave channels
  char *sqName,             // Name of the dma channel in human readable format
  void (*pfCallback) (int, unsigned short, void *),  // Channel callback function (valid only if you are requesting for a DMA MasterChannel)
  int  nTccRequested,       // The TCC that we request to be used (Ignored for Master channels)
  int  *nTccActual,         // Actual TCC used for the PaRAM
  int  *nLch                // Logical Channel Id 0-127 (PaRAM set)
)
{
  int tcc;
  int rc;

  tcc = nTccRequested;

  rc = davinci_request_dma(nDevId, sqName, pfCallback, NULL, nLch, &tcc, EVENTQ_0);

  *nTccActual = tcc;
                           
  if(rc != 0) {
    OSA_EPRINTF(("ERROR davinci_request_dma() failed. rc = %d\n", rc));
    return 0;
  }

  davinci_set_dma_transfer_params(*nLch, 
    4,               // ACNT; 4 since we get REVT for every 32-bits (long word) that we will transmit (McBsp configured in 32-bit mode)
    g_cbBufSize/4,   // BCNT; 128 timeslots in a frame, but we consider the entire buffer a frame (from the DMA point of view)
    1,               // CCNT; this is how many DMA frames. 
    g_cbBufSize/4,   // BCNTRLD; B count reload. the count value used to reload BCNT when the BCNT decrements to 0
    ASYNC
    );

  if (eDirection == ASP_RX_DIR) {
    // ASP_RX_DIR
    davinci_set_dma_dest_params(*nLch, 
      nPhysBufAddr,             // Address to write to
      0,                        // INCR
      0                         // W8BIT (not used in INCR mode)
      );
    davinci_set_dma_src_params(*nLch, 
      (unsigned long)MCBSP_DRR, // Adress to read from
      0,                        // INCR
      0                         // W8BIT (not used in INCR mode)
      );
    davinci_set_dma_src_index(*nLch, 
      0,   // srcbidx; Don't step source. Keep reading from DRR
      0    // srccidx; No third dimension
      );
    davinci_set_dma_dest_index(*nLch, 
      4,   // destbidx; 4 since we need to step buffer source address by 4 bytes for each REVT interrupt (McBsp configured in 32-bit mode)
      0);  // destcidx; No third dimension
  } else {
    // ASP_TX_DIR
    davinci_set_dma_dest_params(*nLch, 
      (unsigned long)MCBSP_DXR, // Address to write to
      0,                        // INCR
      0                         // W8BIT (not used in INCR mode)
      );
    davinci_set_dma_src_params(*nLch, 
      nPhysBufAddr,             // Adress to read from
      0,                        // INCR
      0);                       // W8BIT (not used in INCR mode)
    davinci_set_dma_src_index(*nLch, 
      4,   // srcbidx; 4 since we need to step buffer source address by 4 bytes for each XEVT interrupt (McBsp configured in 32-bit mode)
      0    // srccidx; No third dimension
      );
    davinci_set_dma_dest_index(*nLch, 
      0,   // destbidx; Don't step dest address; keep writing to DXR
      0    // destcidx; No third dimension
      );
  }

  return 1;
}




/******************************************************************************
  OsaLinuxAsp_free_rx_dma
    return 1
******************************************************************************/
static void OsaLinuxAspFreeDmaChannels(void)
{
  // Rx channels
  if (g_nLchMcbspRx != OSA_ASP_INVALID_LCH) {
    davinci_free_dma(g_nLchMcbspRx);
    g_nLchMcbspRx = OSA_ASP_INVALID_LCH;
  }
  if (g_nLchMcbspRxLinkPong != OSA_ASP_INVALID_LCH) {
    davinci_free_dma(g_nLchMcbspRxLinkPong);
    g_nLchMcbspRxLinkPong = OSA_ASP_INVALID_LCH;
  }
  if (g_nLchMcbspRxLinkPing != OSA_ASP_INVALID_LCH) {
    davinci_free_dma(g_nLchMcbspRxLinkPing);
    g_nLchMcbspRxLinkPing = OSA_ASP_INVALID_LCH;
  }

  // Tx channels
  if (g_nLchMcbspTx != OSA_ASP_INVALID_LCH) {
    davinci_free_dma(g_nLchMcbspTx);
    g_nLchMcbspTx = OSA_ASP_INVALID_LCH;
  }
  if (g_nLchMcbspTxLinkPong != OSA_ASP_INVALID_LCH) {
    davinci_free_dma(g_nLchMcbspTxLinkPong);
    g_nLchMcbspTxLinkPong = OSA_ASP_INVALID_LCH;
  }
  if (g_nLchMcbspTxLinkPing != OSA_ASP_INVALID_LCH) {
    davinci_free_dma(g_nLchMcbspTxLinkPing);
    g_nLchMcbspTxLinkPing = OSA_ASP_INVALID_LCH;
  }

}


/******************************************************************************
  OsaLinuxAspOpen
    return 0->OK, Or Error
******************************************************************************/
int OsaLinuxAspOpen(void) 
{
  //int rc = 0;
  u32 nVal;
  int i;

  if (g_eDeviceState != OSA_ASP_STATE_INITIALIZED) {
    OSA_DPRINTF(("OsaLinuxAspOpen: device is not in the initalized state\n"));
    return -EEXIST;
  }

  davinci_mcbsp_stop_rx(ASP_MCBSP);
  davinci_mcbsp_stop_tx(ASP_MCBSP);

  davinci_mcbsp_config(ASP_MCBSP, &g_sMcBspInitialConfig);

  i = davinci_mcbsp_request(ASP_MCBSP);
  if(i != 0) {
    if(i == -1) {
      OSA_DPRINTF(("devasp: MCBSP1 in use trying to free\n"));
      davinci_mcbsp_free(ASP_MCBSP);
      i = davinci_mcbsp_request(ASP_MCBSP);
      if(i != 0) {
        OSA_EPRINTF(("devasp: can't claim MCBSP1, code %d\n", i));
        return -EEXIST;
      }
    }
    else {
      OSA_EPRINTF(("devasp: can't claim MCBSP1, code %d\n", i));
      return -EEXIST;
    }
  }

  g_nLchMcbspRx         = OSA_ASP_INVALID_LCH; // Master channel Rx
  g_nLchMcbspRxLinkPong = OSA_ASP_INVALID_LCH;
  g_nLchMcbspRxLinkPing = OSA_ASP_INVALID_LCH;
  g_nLchMcbspTx         = OSA_ASP_INVALID_LCH; // Master channel Rx
  g_nLchMcbspTxLinkPong = OSA_ASP_INVALID_LCH;
  g_nLchMcbspTxLinkPing = OSA_ASP_INVALID_LCH;

  g_bRxPingBufReady = 0;  // Boolean, If true then we ping buf is ready to be accessed. If false, the buffer is not ready to be accessed
  g_bRxPongBufReady = 0;  // Boolean, If true then we pong buf is ready to be accessed. If false, the buffer is not ready to be accessed
  g_bTxPingBufReady = 0;  // Boolean, If true then we ping buf is ready to be accessed. If false, the buffer is not ready to be accessed
  g_bTxPongBufReady = 0;  // Boolean, If true then we pong buf is ready to be accessed. If false, the buffer is not ready to be accessed


  // Lower the MEWC extended wait cycles. The EMIF will wait for a maximum of (MEWC + 1)*16 clock cycles before it stops inserting asynchronous wait cycles and proceeds to the hold period of the access.
  OSA_DPRINTF((" DAVINCI_AWCCR  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_AWCCR)));    // Asynchronous 1 Configuration Register (CS2 Space)
  nVal = OsaLinuxAspReadReg32(DAVINCI_AWCCR); 
  g_nSavedAwccr = nVal;
  nVal &= ~0x000000FF; // Mask out MEWC bits
  nVal |= 0x10;        // Set MEWC to a lower value of 0x10 which is (16+1)*16=272 99MHz (10ns) cycles. That equates to 272 * 10ns = 2.72us timeout value. This is less than the 4us time that we have to service the MsBsp (assuming 32-bit words and 8MHz clock rate).
  OsaLinuxAspWriteReg32(DAVINCI_AWCCR, nVal); 
  nVal = OsaLinuxAspReadReg32(DAVINCI_AWCCR); 
  OSA_DPRINTF((" DAVINCI_AWCCR  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_AWCCR)));    // Asynchronous 1 Configuration Register (CS2 Space)

  // Set ASP interrupt priority (This might not be necessary)
  nVal = OsaLinuxAspReadReg32(DAVINCI_INTPRI3);  // INTPRI3
  //OSA_DPRINTF(("INTPRI3 = 0x%.8X\n", nVal));
  nVal &= ~0x00000077; // Mask out the ASP priority interrupts bits (interrupt 24 and 25)
  nVal |=  0x00000022; // Set interrupt to 2 which is the highest non-FIQ setting.
  OsaLinuxAspWriteReg32(DAVINCI_INTPRI3, nVal);  // INTPRI3
  nVal = OsaLinuxAspReadReg32(DAVINCI_INTPRI3);  // INTPRI3
  //OSA_DPRINTF(("INTPRI3 = 0x%.8X\n", nVal));

  g_eDeviceState = OSA_ASP_STATE_OPENED;

  if(!OsaLinuxAspAllocBuffers(g_cbBufSize)) goto fail;

  return 0;

fail:
  OSA_EPRINTF(("OsaLinuxAsp_open: FAILED\n"));

  OsaLinuxAspFreeBuffers();

  g_eDeviceState = OSA_ASP_STATE_INITIALIZED; // back to the previous state

  return -ENOMEM;
}



/******************************************************************************
  OsaLinuxCreateAndConfigureDmaChannels
    return 0 on success, negative value on Error
******************************************************************************/
static int OsaLinuxCreateAndConfigureDmaChannels(void) 
{
  int rc = 0;

  // ************* RX CHANNELS **************************

  // Create Master RX channel (head channel) (PING)
  rc = OsaLinuxAspCreateDmaChannel(
    ASP_RX_DIR,                               // ASP_RX_DIR, or ASP_TX_DIR
    g_sqDmaBuffers[OSA_ASP_BUF_RX_PING].phys,  // Physical address of the buffer for this channel
    DAVINCI_DMA_MCBSP1_RX,                    // DMA channel number. DAVINCI_DMA_MCBSP1_RX for master or DAVINCI_EDMA_PARAM_ANY for slave channels
    "MCBSP RX MASTER",                        // Name of the dma channel in human readable format
    OsaLinuxAspIrqRxMaster,                   // Channel callback function (valid only if you are requesting for a DMA MasterChannel)
    3,                                        // The TCC that we request to be used (Ignored for Master channels)
    &g_nTccMcbspRx,                           // Actual TCC used for the PaRAM
    &g_nLchMcbspRx                            // Logical Channel Id 0-127 (PaRAM set)
  );
  if (!rc) goto fail;

  //OSA_DPRINTF(("lch g_nLchMcbspRx = %d, tcc=%d (not used)\n", g_nLchMcbspRx, g_nTccMcbspRx));

  // Create Pong Slave Link RX channel 
  rc = OsaLinuxAspCreateDmaChannel(
    ASP_RX_DIR,                               // ASP_RX_DIR, or ASP_TX_DIR
    g_sqDmaBuffers[OSA_ASP_BUF_RX_PONG].phys,  // Physical address of the buffer for this channel
    DAVINCI_DMA_CHANNEL_ANY/*DAVINCI_EDMA_PARAM_ANY*/,                   // DMA channel number. DAVINCI_DMA_MCBSP1_RX for master or DAVINCI_EDMA_PARAM_ANY for slave channels
    "MCBSP RX LINK PONG",                     // Name of the dma channel in human readable format
    OsaLinuxAspIrqRxPong,                                     // Channel callback function (valid only if you are requesting for a DMA MasterChannel)
    OSA_ASP_TCC_RX_LINK_PONG,                  // The TCC that we request to be used (Ignored for Master channels)
    &g_nTccMcbspRxLinkPong,                   // Actual TCC used for the PaRAM
    &g_nLchMcbspRxLinkPong                    // Logical Channel Id 0-127 (PaRAM set)
  );
  if (!rc) goto fail;

  //OSA_DPRINTF(("lch g_nLchMcbspRxLinkPong = %d, tcc=%d\n", g_nLchMcbspRxLinkPong, g_nTccMcbspRxLinkPong));

  // Create Ping Slave Link RX channel 
  rc = OsaLinuxAspCreateDmaChannel(
    ASP_RX_DIR,                               // ASP_RX_DIR, or ASP_TX_DIR
    g_sqDmaBuffers[OSA_ASP_BUF_RX_PING].phys,  // Physical address of the buffer for this channel
    DAVINCI_DMA_CHANNEL_ANY/*DAVINCI_EDMA_PARAM_ANY*/,                   // DMA channel number. DAVINCI_DMA_MCBSP1_RX for master or DAVINCI_EDMA_PARAM_ANY for slave channels
    "MCBSP RX LINK PING",                     // Name of the dma channel in human readable format
    OsaLinuxAspIrqRxPing,                                     // Channel callback function (valid only if you are requesting for a DMA MasterChannel)
    OSA_ASP_TCC_RX_LINK_PONG/*OSA_ASP_TCC_RX_LINK_PING*/,                  // The TCC that we request to be used (Ignored for Master channels)
    &g_nTccMcbspRxLinkPing,                   // Actual TCC used for the PaRAM
    &g_nLchMcbspRxLinkPing                    // Logical Channel Id 0-127 (PaRAM set)
  );
  if (!rc) goto fail;

  //OSA_DPRINTF(("lch g_nLchMcbspRxLinkPing = %d, tcc=%d\n", g_nLchMcbspRxLinkPing, g_nTccMcbspRxLinkPing));

  // Link the master to the slave and then the slaves back to each other (pong/ping)
  davinci_dma_link_lch(g_nLchMcbspRx, g_nLchMcbspRxLinkPong);
  davinci_dma_link_lch(g_nLchMcbspRxLinkPong, g_nLchMcbspRxLinkPing);
  davinci_dma_link_lch(g_nLchMcbspRxLinkPing, g_nLchMcbspRxLinkPong);


  // ************* TX CHANNELS **************************

  // Create Master TX channel (head channel) (PING)
  rc = OsaLinuxAspCreateDmaChannel(
    ASP_TX_DIR,                               // ASP_TX_DIR, or ASP_TX_DIR
    g_sqDmaBuffers[OSA_ASP_BUF_TX_PING].phys,  // Physical address of the buffer for this channel
    DAVINCI_DMA_MCBSP1_TX,                    // DMA channel number. DAVINCI_DMA_MCBSP1_TX for master or DAVINCI_EDMA_PARAM_ANY for slave channels
    "MCBSP TX MASTER",                        // Name of the dma channel in human readable format
    OsaLinuxAspIrqTxMaster,                  // Channel callback function (valid only if you are requesting for a DMA MasterChannel)
    2,                                        // The TCC that we request to be used (Ignored for Master channels)
    &g_nTccMcbspTx,                           // Actual TCC used for the PaRAM
    &g_nLchMcbspTx                            // Logical Channel Id 0-127 (PaRAM set)
  );
  if (!rc) goto fail;

  //OSA_DPRINTF(("lch g_nLchMcbspTx = %d, tcc=%d (not used)\n", g_nLchMcbspTx, g_nTccMcbspTx));

  // Create Pong Slave Link TX channel 
  rc = OsaLinuxAspCreateDmaChannel(
    ASP_TX_DIR,                               // ASP_TX_DIR, or ASP_TX_DIR
    g_sqDmaBuffers[OSA_ASP_BUF_TX_PONG].phys,  // Physical address of the buffer for this channel
    DAVINCI_DMA_CHANNEL_ANY/*DAVINCI_EDMA_PARAM_ANY*/,                   // DMA channel number. DAVINCI_DMA_MCBSP1_TX for master or DAVINCI_EDMA_PARAM_ANY for slave channels
    "MCBSP TX LINK PONG",                     // Name of the dma channel in human readable format
    OsaLinuxAspIrqTxPong,                                     // Channel callback function (valid only if you are requesting for a DMA MasterChannel)
    OSA_ASP_TCC_TX_LINK_PONG,                  // The TCC that we request to be used (Ignored for Master channels)
    &g_nTccMcbspTxLinkPong,                   // Actual TCC used for the PaRAM
    &g_nLchMcbspTxLinkPong                    // Logical Channel Id 0-127 (PaRAM set)
  );
  if (!rc) goto fail;

  //OSA_DPRINTF(("lch g_nLchMcbspTxLinkPong = %d, tcc=%d\n", g_nLchMcbspTxLinkPong, g_nTccMcbspTxLinkPong));

  // Create Ping Slave Link TX channel 
  rc = OsaLinuxAspCreateDmaChannel(
    ASP_TX_DIR,                               // ASP_TX_DIR, or ASP_TX_DIR
    g_sqDmaBuffers[OSA_ASP_BUF_TX_PING].phys,  // Physical address of the buffer for this channel
    DAVINCI_DMA_CHANNEL_ANY/*DAVINCI_EDMA_PARAM_ANY*/,                   // DMA channel number. DAVINCI_DMA_MCBSP1_TX for master or DAVINCI_EDMA_PARAM_ANY for slave channels
    "MCBSP TX LINK PING",                     // Name of the dma channel in human readable format
    OsaLinuxAspIrqTxPing,                                     // Channel callback function (valid only if you are requesting for a DMA MasterChannel)
    OSA_ASP_TCC_TX_LINK_PONG/*OSA_ASP_TCC_TX_LINK_PING*/,                  // The TCC that we request to be used (Ignored for Master channels)
    &g_nTccMcbspTxLinkPing,                   // Actual TCC used for the PaRAM
    &g_nLchMcbspTxLinkPing                    // Logical Channel Id 0-127 (PaRAM set)
  );
  if (!rc) goto fail;

  //OSA_DPRINTF(("lch g_nLchMcbspTxLinkPing = %d, tcc=%d\n", g_nLchMcbspTxLinkPing, g_nTccMcbspTxLinkPing));

  // Link the master to the slave and then the slaves back to each other (pong/ping)
  davinci_dma_link_lch(g_nLchMcbspTx, g_nLchMcbspTxLinkPong);
  davinci_dma_link_lch(g_nLchMcbspTxLinkPong, g_nLchMcbspTxLinkPing);
  davinci_dma_link_lch(g_nLchMcbspTxLinkPing, g_nLchMcbspTxLinkPong);

#if 0
  {
    edmacc_paramentry_regs r;
    davinci_get_dma_params(g_nLchMcbspRx, &r);
    OSA_DPRINTF(("RX Master edmacc_paramentry_regs:\n"));
    OSA_DPRINTF(("  opt          = 0x%.8X\n", r.opt));
    OSA_DPRINTF(("  src          = 0x%.8X\n", r.src));
    OSA_DPRINTF(("  a_b_cnt      = 0x%.8X\n", r.a_b_cnt));
    OSA_DPRINTF(("  dst          = 0x%.8X\n", r.dst));
    OSA_DPRINTF(("  src_dst_bidx = 0x%.8X\n", r.src_dst_bidx));
    OSA_DPRINTF(("  link_bcntrld = 0x%.8X\n", r.link_bcntrld));
    OSA_DPRINTF(("  src_dst_cidx = 0x%.8X\n", r.src_dst_cidx));
    OSA_DPRINTF(("  ccnt         = 0x%.8X\n", r.ccnt));
    davinci_get_dma_params(g_nLchMcbspRxLinkPong, &r);
    OSA_DPRINTF(("RX Pong edmacc_paramentry_regs:\n"));
    OSA_DPRINTF(("  opt          = 0x%.8X\n", r.opt));
    OSA_DPRINTF(("  src          = 0x%.8X\n", r.src));
    OSA_DPRINTF(("  a_b_cnt      = 0x%.8X\n", r.a_b_cnt));
    OSA_DPRINTF(("  dst          = 0x%.8X\n", r.dst));
    OSA_DPRINTF(("  src_dst_bidx = 0x%.8X\n", r.src_dst_bidx));
    OSA_DPRINTF(("  link_bcntrld = 0x%.8X\n", r.link_bcntrld));
    OSA_DPRINTF(("  src_dst_cidx = 0x%.8X\n", r.src_dst_cidx));
    OSA_DPRINTF(("  ccnt         = 0x%.8X\n", r.ccnt));
    davinci_get_dma_params(g_nLchMcbspRxLinkPing, &r);
    OSA_DPRINTF(("RX Ping edmacc_paramentry_regs:\n"));
    OSA_DPRINTF(("  opt          = 0x%.8X\n", r.opt));
    OSA_DPRINTF(("  src          = 0x%.8X\n", r.src));
    OSA_DPRINTF(("  a_b_cnt      = 0x%.8X\n", r.a_b_cnt));
    OSA_DPRINTF(("  dst          = 0x%.8X\n", r.dst));
    OSA_DPRINTF(("  src_dst_bidx = 0x%.8X\n", r.src_dst_bidx));
    OSA_DPRINTF(("  link_bcntrld = 0x%.8X\n", r.link_bcntrld));
    OSA_DPRINTF(("  src_dst_cidx = 0x%.8X\n", r.src_dst_cidx));
    OSA_DPRINTF(("  ccnt         = 0x%.8X\n", r.ccnt));
    davinci_get_dma_params(g_nLchMcbspTx, &r);
    OSA_DPRINTF(("TX Master edmacc_paramentry_regs:\n"));
    OSA_DPRINTF(("  opt          = 0x%.8X\n", r.opt));
    OSA_DPRINTF(("  src          = 0x%.8X\n", r.src));
    OSA_DPRINTF(("  a_b_cnt      = 0x%.8X\n", r.a_b_cnt));
    OSA_DPRINTF(("  dst          = 0x%.8X\n", r.dst));
    OSA_DPRINTF(("  src_dst_bidx = 0x%.8X\n", r.src_dst_bidx));
    OSA_DPRINTF(("  link_bcntrld = 0x%.8X\n", r.link_bcntrld));
    OSA_DPRINTF(("  src_dst_cidx = 0x%.8X\n", r.src_dst_cidx));
    OSA_DPRINTF(("  ccnt         = 0x%.8X\n", r.ccnt));
    davinci_get_dma_params(g_nLchMcbspTxLinkPong, &r);
    OSA_DPRINTF(("TX Pong Link edmacc_paramentry_regs:\n"));
    OSA_DPRINTF(("  opt          = 0x%.8X\n", r.opt));
    OSA_DPRINTF(("  src          = 0x%.8X\n", r.src));
    OSA_DPRINTF(("  a_b_cnt      = 0x%.8X\n", r.a_b_cnt));
    OSA_DPRINTF(("  dst          = 0x%.8X\n", r.dst));
    OSA_DPRINTF(("  src_dst_bidx = 0x%.8X\n", r.src_dst_bidx));
    OSA_DPRINTF(("  link_bcntrld = 0x%.8X\n", r.link_bcntrld));
    OSA_DPRINTF(("  src_dst_cidx = 0x%.8X\n", r.src_dst_cidx));
    OSA_DPRINTF(("  ccnt         = 0x%.8X\n", r.ccnt));
    davinci_get_dma_params(g_nLchMcbspTxLinkPing, &r);
    OSA_DPRINTF(("TX Ping Link edmacc_paramentry_regs\n"));
    OSA_DPRINTF(("  opt          = 0x%.8X\n", r.opt));
    OSA_DPRINTF(("  src          = 0x%.8X\n", r.src));
    OSA_DPRINTF(("  a_b_cnt      = 0x%.8X\n", r.a_b_cnt));
    OSA_DPRINTF(("  dst          = 0x%.8X\n", r.dst));
    OSA_DPRINTF(("  src_dst_bidx = 0x%.8X\n", r.src_dst_bidx));
    OSA_DPRINTF(("  link_bcntrld = 0x%.8X\n", r.link_bcntrld));
    OSA_DPRINTF(("  src_dst_cidx = 0x%.8X\n", r.src_dst_cidx));
    OSA_DPRINTF(("  ccnt         = 0x%.8X\n", r.ccnt));
  }
#endif
#if 0
  OSA_DPRINTF(("Other important DMA registers\n"));
  OSA_DPRINTF(("  er          = 0x%.8X\n", ptr_edmacc_regs->er ));
  OSA_DPRINTF(("  erh         = 0x%.8X\n", ptr_edmacc_regs->erh ));
  OSA_DPRINTF(("  ser         = 0x%.8X\n", ptr_edmacc_regs->ser ));
  OSA_DPRINTF(("  serh        = 0x%.8X\n", ptr_edmacc_regs->serh ));
  OSA_DPRINTF(("  emr         = 0x%.8X\n", ptr_edmacc_regs->emr ));
  OSA_DPRINTF(("  emrh        = 0x%.8X\n", ptr_edmacc_regs->emrh ));
  OSA_DPRINTF(("  ipr         = 0x%.8X\n", ptr_edmacc_regs->ipr ));
  OSA_DPRINTF(("  iprh        = 0x%.8X\n", ptr_edmacc_regs->iprh ));
  OSA_DPRINTF(("  ier         = 0x%.8X\n", ptr_edmacc_regs->ier ));
  OSA_DPRINTF(("  ierh        = 0x%.8X\n", ptr_edmacc_regs->ierh ));
  OSA_DPRINTF(("  eer         = 0x%.8X\n", ptr_edmacc_regs->eer ));
  OSA_DPRINTF(("  eerh        = 0x%.8X\n", ptr_edmacc_regs->eerh ));
#endif

#if 0
  {
    int i;
    OSA_DPRINTF(("Complete edmacc_paramentry_regs dump:\n"));
    for (i = 0; i < 128; i++) {
      OSA_DPRINTF(("[%d] address = 0x%p \n", i, &(ptr_edmacc_regs->paramentry[i])));
      OSA_DPRINTF(("  opt          = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].opt ));
      OSA_DPRINTF(("  src          = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].src ));
      OSA_DPRINTF(("  a_b_cnt      = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].a_b_cnt ));
      OSA_DPRINTF(("  dst          = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].dst ));
      OSA_DPRINTF(("  src_dst_bidx = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].src_dst_bidx ));
      OSA_DPRINTF(("  link_bcntrld = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].link_bcntrld ));
      OSA_DPRINTF(("  src_dst_cidx = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].src_dst_cidx ));
      OSA_DPRINTF(("  ccnt         = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].ccnt ));
    }
  }
#endif

  return 0; // Success

fail:
  OSA_EPRINTF(("OsaLinuxCreateAndConfigureDmaChannels: FAILED\n"));

  OsaLinuxAspFreeDmaChannels();

  return -ENOMEM;
}


/******************************************************************************
  OsaLinuxAspMcbspStart

  Here we start the McBSP receiver, by enabling the sample
  generator, both transmitter and receivers, and the frame sync.

  return void
******************************************************************************/
static void OsaLinuxAspMcbspStart(int bStartRx, int bStartTx)
{
	u32 w;


	w = OsaLinuxAspReadReg32(MCBSP_SPCR);
	OsaLinuxAspWriteReg32(MCBSP_SPCR, w |
    MCBSP_SPCR_GRST |  // Setting GRST to take the sample Generator out of reset
    0);

	udelay(100); // Wait a while. I guess to let the clocks settle.

  // Prime the transmitter (dummy value)
  OsaLinuxAspWriteReg32(MCBSP_DXR, 0x55); // Arbitrary value

  // Start the receiver and transmitter simultaneously
	w = OsaLinuxAspReadReg32(MCBSP_SPCR);
	OsaLinuxAspWriteReg32(MCBSP_SPCR, w |
    ((bStartRx) ? MCBSP_SPCR_RRST : 0) |  // Setting RRST to take the Receiver out of reset
    ((bStartTx) ? MCBSP_SPCR_XRST : 0) |  // Setting XRST to take the Transmitter out of reset
    MCBSP_SPCR_FRST |  // Setting FRST to start Frame sync generation
    0);


#if 0
	w = OsaLinuxAspReadReg32(MCBSP_SPCR);
  OSA_DPRINTF(("SPCR = 0x%.8X\n", w));

  w = OsaLinuxAspReadReg32(0x01C48008);
  OSA_DPRINTF(("IRQ0 = 0x%.8X\n", w));

  w = OsaLinuxAspReadReg32(0x01C4803C);  // INTPRI3
  OSA_DPRINTF(("INTPRI3 = 0x%.8X\n", w));

  w = OsaLinuxAspReadReg32(0x01C48000);
  OSA_DPRINTF(("FIQ0 = 0x%.8X\n", w));

  w = OsaLinuxAspReadReg32(0x01C48018);
  OSA_DPRINTF(("EINT0 = 0x%.8X\n", w));
  

  OSA_DPRINTF(("  er          = 0x%.8X\n", ptr_edmacc_regs->er ));
  OSA_DPRINTF(("  emr         = 0x%.8X\n", ptr_edmacc_regs->emr ));
  OSA_DPRINTF(("  ipr         = 0x%.8X\n", ptr_edmacc_regs->ipr ));
  OSA_DPRINTF(("  ier         = 0x%.8X\n", ptr_edmacc_regs->ier ));
  OSA_DPRINTF(("  eer         = 0x%.8X\n", ptr_edmacc_regs->eer ));
#endif
}


/******************************************************************************
  OsaLinuxAspEnable
    return 0->OK, Or Error
******************************************************************************/
int OsaLinuxAspEnable(int bStartRx, int bStartTx) 
{

  if (g_eDeviceState != OSA_ASP_STATE_OPENED){
    OSA_DPRINTF(("OsaLinuxAspEnable: device is not in its opened state\n"));
    return -EEXIST;
  }

  if (OsaLinuxCreateAndConfigureDmaChannels()) {
    OSA_DPRINTF(("OsaLinuxAspEnable: Could not create and configure the DMA channels\n"));
    return -EEXIST;
  }

  davinci_mcbsp_stop_rx(ASP_MCBSP);
  davinci_mcbsp_stop_tx(ASP_MCBSP);

  davinci_clean_channel(g_nLchMcbspRx);
  davinci_clean_channel(g_nLchMcbspTx);

  if (bStartRx) {
    davinci_start_dma(g_nLchMcbspRx);
  }
  if (bStartTx) {
    davinci_start_dma(g_nLchMcbspTx);
  }

  OsaLinuxAspMcbspStart(bStartRx, bStartTx);   // Neal said that we should start the McBSP before the DMA. However, I don't see how we can get in sync if we do it that way. Therefore we start the McBSP after we start the DMA.

  g_eDeviceState = OSA_ASP_STATE_ENABLED; // Everything is running

  return 0; // Success

}

/******************************************************************************
  OsaLinuxAspEnable
    return 0->OK, Or Error
******************************************************************************/
int OsaLinuxAspDisable(void) 
{
  int i;

  if (g_eDeviceState != OSA_ASP_STATE_ENABLED){
    OSA_DPRINTF(("OsaLinuxAspDisable: device is not in its enabled state\n"));
    return -EEXIST;
  }

  davinci_stop_dma(g_nLchMcbspRx);
  davinci_stop_dma(g_nLchMcbspRxLinkPong);
  davinci_stop_dma(g_nLchMcbspRxLinkPing);

  davinci_clean_channel(g_nLchMcbspRx);
  davinci_clean_channel(g_nLchMcbspRxLinkPong);
  davinci_clean_channel(g_nLchMcbspRxLinkPing);
  
  davinci_stop_dma(g_nLchMcbspTx);
  davinci_stop_dma(g_nLchMcbspTxLinkPong);
  davinci_stop_dma(g_nLchMcbspTxLinkPing);

  davinci_clean_channel(g_nLchMcbspTx);
  davinci_clean_channel(g_nLchMcbspTxLinkPong);
  davinci_clean_channel(g_nLchMcbspTxLinkPing);

  // Wait for any pending DMAs to finish (the maximum value for udelay is 2000us so we loop here a number of times so we can delay our 8000 ms
  for (i = 0; i < (g_cbBufSize/128 * 2); i++) {   // Loop for g_cbBufSize/128 is how many frames. Then times two just to be safe
    udelay(125); // 125us is the period of one frame. 
  }

  davinci_mcbsp_stop_rx(ASP_MCBSP);
  davinci_mcbsp_stop_tx(ASP_MCBSP);

  OsaLinuxAspFreeDmaChannels();

  g_eDeviceState = OSA_ASP_STATE_OPENED; // Back to the previous state

  return 0;

}



/******************************************************************************
  OsaLinuxAspRead
    return count, Or Error if negative
******************************************************************************/
ssize_t OsaLinuxAspRead(char *buffer, size_t count)
{
  if (g_eDeviceState == OSA_ASP_STATE_ENABLED) {
    if (g_bRxPingBufReady) {
      memcpy(buffer, g_sqDmaBuffers[OSA_ASP_BUF_RX_PING].virt, count);
      g_bRxPingBufReady = 0; // Indicate that we are done with this buffer
    } else if (g_bRxPongBufReady) {
      memcpy(buffer, g_sqDmaBuffers[OSA_ASP_BUF_RX_PONG].virt, count);
      g_bRxPongBufReady = 0; // Indicate that we are done with this buffer
    } else {
      return -ENOMEM; // No buffers are available
    }
    return count; // Success
  } else {
    return -ENOMEM; // The device was not enabled
  }
}

/******************************************************************************
  OsaLinuxAspGetReadPtr
    return 1 on success, Or Error if negative
******************************************************************************/
ssize_t OsaLinuxAspGetReadPtr(char **pBuffer)
{
  if (g_eDeviceState == OSA_ASP_STATE_ENABLED) {
    if (g_bRxPingBufReady) {
      *pBuffer = g_sqDmaBuffers[OSA_ASP_BUF_RX_PING].virt;
      g_bRxPingBufReady = 0; // Indicate that we are done with this buffer
    } else if (g_bRxPongBufReady) {
      *pBuffer = g_sqDmaBuffers[OSA_ASP_BUF_RX_PONG].virt;
      g_bRxPongBufReady = 0; // Indicate that we are done with this buffer
    } else {
      return -ENOMEM; // No buffers are available
    }
    return 1; // Success
  } else {
    return -ENOMEM; // The device was not enabled
  }
}



/******************************************************************************
  OsaLinuxAspWrite
    return count, Or Error if negative
******************************************************************************/
ssize_t OsaLinuxAspWrite(const char *buffer, size_t count)
{
  if (g_eDeviceState == OSA_ASP_STATE_ENABLED) {
    if (g_bTxPingBufReady) {
      memcpy(g_sqDmaBuffers[OSA_ASP_BUF_TX_PING].virt, buffer, count);
      g_bTxPingBufReady = 0; // Indicate that we are done with this buffer
    } else if (g_bTxPongBufReady) {
      memcpy(g_sqDmaBuffers[OSA_ASP_BUF_TX_PONG].virt, buffer, count);
      g_bTxPongBufReady = 0; // Indicate that we are done with this buffer
    } else {
      return -ENOMEM; // No buffers are available
    }
    return count; // Success
  } else {
    return -ENOMEM; // The device was not enabled
  }
}


/******************************************************************************
  OsaLinuxAspGetBufferPointers

    The user app can read/write directly to/from the ping and pong buffers. This function
    will provide the pointers to the buffer that he would need to access.

    return 0->OK, Or Error
******************************************************************************/
int OsaLinuxAspGetBufferPointers(char **pRxPing, char **pRxPong, char **pTxPing, char **pTxPong)
{
  if (g_eDeviceState != OSA_ASP_STATE_OPENED) {
    OSA_EPRINTF(("OsaLinuxAspGetBufferPointers: The buffer pointers can not be retrieved in this state.\n"));
    return -EINVAL;
  }

  *pRxPing = g_sqDmaBuffers[OSA_ASP_BUF_RX_PING].virt;
  *pRxPong = g_sqDmaBuffers[OSA_ASP_BUF_RX_PONG].virt;
  *pTxPing = g_sqDmaBuffers[OSA_ASP_BUF_TX_PING].virt;
  *pTxPong = g_sqDmaBuffers[OSA_ASP_BUF_TX_PONG].virt;
  return 0;
}


/******************************************************************************
******************************************************************************/
int OsaLinuxSetAspCfg(OsaLinuxAspRegCfgS *pRegCfg)
{
  memcpy(&g_sMcBspInitialConfig, pRegCfg, sizeof(OsaLinuxAspRegCfgS));
  return 0; // Success
}


/******************************************************************************
  OsaLinuxAspInitDevAsp
    return 0->OK, Or Error
******************************************************************************/
int OsaLinuxAspInit(int cbBufSize, void *pContext, void (*funcServeTxIrq)(unsigned long, void *), void (*funcServeRxIrq)(unsigned long, void *))
{
  int i;

  OSA_DPRINTF(("OsaLinuxAspInitDevAsp() +++++++++++++++ g_cbBufSize=%d\n", g_cbBufSize));

  if (g_eDeviceState != OSA_ASP_STATE_UNINITIALIZED) {
    goto fail;
  }

  g_cbBufSize = cbBufSize;
  g_pCallbackContext   = pContext;
  g_pCallbackFuncTxIrq = funcServeTxIrq;
  g_pCallbackFuncRxIrq = funcServeRxIrq;


  // Initialize Buffer pointers to NULL. They will be allocated later in OsaLinuxAspOpen()
  for(i = 0; i < OSA_ASP_N_BUFFERS; i++) {
    g_sqDmaBuffers[i].virt = NULL;
    g_sqDmaBuffers[i].phys = 0;
  }


  ptr_edmacc_regs = ((edmacc_regs *) IO_ADDRESS(DAVINCI_DMA_REGISTER_BASE));
	OSA_DPRINTF(("DMA REG BASE ADDR=%p\n", ptr_edmacc_regs));

  OSA_DPRINTF(("address paramentry[0] = 0x%X \n", (u32)&ptr_edmacc_regs->paramentry[0]));

#if 0
  {
    int i;
    OSA_DPRINTF(("Complete edmacc_paramentry_regs dump:\n"));
    for (i = 0; i < 128; i++) {
      OSA_DPRINTF(("[%d] address = 0x%p \n", i, &(ptr_edmacc_regs->paramentry[i])));
      OSA_DPRINTF(("  opt          = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].opt ));
      OSA_DPRINTF(("  src          = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].src ));
      OSA_DPRINTF(("  a_b_cnt      = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].a_b_cnt ));
      OSA_DPRINTF(("  dst          = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].dst ));
      OSA_DPRINTF(("  src_dst_bidx = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].src_dst_bidx ));
      OSA_DPRINTF(("  link_bcntrld = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].link_bcntrld ));
      OSA_DPRINTF(("  src_dst_cidx = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].src_dst_cidx ));
      OSA_DPRINTF(("  ccnt         = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].ccnt ));
    }
  }
#endif

#if 0
  {
    OSA_DPRINTF(("Other important DMA registers\n"));
    OSA_DPRINTF(("  quepri      = 0x%.8X\n", ptr_edmacc_regs->quepri ));
    OSA_DPRINTF(("  er          = 0x%.8X\n", ptr_edmacc_regs->er ));
    OSA_DPRINTF(("  erh         = 0x%.8X\n", ptr_edmacc_regs->erh ));
    OSA_DPRINTF(("  ser         = 0x%.8X\n", ptr_edmacc_regs->ser ));
    OSA_DPRINTF(("  serh        = 0x%.8X\n", ptr_edmacc_regs->serh ));
    OSA_DPRINTF(("  emr         = 0x%.8X\n", ptr_edmacc_regs->emr ));
    OSA_DPRINTF(("  emrh        = 0x%.8X\n", ptr_edmacc_regs->emrh ));
    OSA_DPRINTF(("  ipr         = 0x%.8X\n", ptr_edmacc_regs->ipr ));
    OSA_DPRINTF(("  iprh        = 0x%.8X\n", ptr_edmacc_regs->iprh ));
    OSA_DPRINTF(("  ier         = 0x%.8X\n", ptr_edmacc_regs->ier ));
    OSA_DPRINTF(("  ierh        = 0x%.8X\n", ptr_edmacc_regs->ierh ));
    OSA_DPRINTF(("  eer         = 0x%.8X\n", ptr_edmacc_regs->eer ));
    OSA_DPRINTF(("  eerh        = 0x%.8X\n", ptr_edmacc_regs->eerh ));
    
    OSA_DPRINTF(("Switched Central Resource registers\n"));
    OSA_DPRINTF((" DAVINCI_MSTPRI0  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_MSTPRI0)));  
    OSA_DPRINTF((" DAVINCI_MSTPRI1  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_MSTPRI1)));  
  }
#endif

  g_eDeviceState = OSA_ASP_STATE_INITIALIZED;

  return 0;

fail:
  OSA_EPRINTF(("OsaLinuxAspInit: Failed\n"));
  return -EINVAL;

}


/******************************************************************************
  OsaLinuxAspClose
    return void
******************************************************************************/
void OsaLinuxAspClose(void)
{

  if (g_eDeviceState == OSA_ASP_STATE_ENABLED) {
    OsaLinuxAspDisable(); // Changes the state to OSA_ASP_STATE_OPENED
  }

  if (g_eDeviceState == OSA_ASP_STATE_OPENED) {

    OsaLinuxAspFreeBuffers();
    
    davinci_mcbsp_free(ASP_MCBSP);

    g_eDeviceState = OSA_ASP_STATE_INITIALIZED;

    // Restore the AWCCR value so that we can access the NAND flash normally
    OsaLinuxAspWriteReg32(DAVINCI_AWCCR, g_nSavedAwccr); 

  }


}


/******************************************************************************
  OsaLinuxAspFreeBuffers
    return void
******************************************************************************/
static void OsaLinuxAspFreeBuffers(void)
{
  int i;

  for(i = 0; i < OSA_ASP_N_BUFFERS; i++) {
    if(g_sqDmaBuffers[i].virt != NULL) {
#ifdef OSA_ASP_USE_DDR_MEM_FOR_BUFFERS
      dma_free_coherent(NULL, g_cbBufSize, g_sqDmaBuffers[i].virt, g_sqDmaBuffers[i].phys);
#else // Use VICP memory
      iounmap(g_sqDmaBuffers[i].virt);
#endif
    }
  }
}


/******************************************************************************
  OsaLinuxAspDbgDump (Debug function)
******************************************************************************/
int OsaLinuxAspDbgDump(OsaUint32 nCmd)
{

  if (nCmd == 1) {
    OSA_DPRINTF(("Important DMA registers\n"));
    OSA_DPRINTF(("  quepri      = 0x%.8X\n", ptr_edmacc_regs->quepri ));
    OSA_DPRINTF(("  er          = 0x%.8X\n", ptr_edmacc_regs->er ));
    OSA_DPRINTF(("  erh         = 0x%.8X\n", ptr_edmacc_regs->erh ));
    OSA_DPRINTF(("  ser         = 0x%.8X\n", ptr_edmacc_regs->ser ));
    OSA_DPRINTF(("  serh        = 0x%.8X\n", ptr_edmacc_regs->serh ));
    OSA_DPRINTF(("  emr         = 0x%.8X\n", ptr_edmacc_regs->emr ));
    OSA_DPRINTF(("  emrh        = 0x%.8X\n", ptr_edmacc_regs->emrh ));
    OSA_DPRINTF(("  ipr         = 0x%.8X\n", ptr_edmacc_regs->ipr ));
    OSA_DPRINTF(("  iprh        = 0x%.8X\n", ptr_edmacc_regs->iprh ));
    OSA_DPRINTF(("  ier         = 0x%.8X\n", ptr_edmacc_regs->ier ));
    OSA_DPRINTF(("  ierh        = 0x%.8X\n", ptr_edmacc_regs->ierh ));
    OSA_DPRINTF(("  eer         = 0x%.8X\n", ptr_edmacc_regs->eer ));
    OSA_DPRINTF(("  eerh        = 0x%.8X\n", ptr_edmacc_regs->eerh ));
    OSA_DPRINTF(("  ipr         = 0x%.8X\n", ptr_edmacc_regs->ipr ));
    OSA_DPRINTF(("  iprh        = 0x%.8X\n", ptr_edmacc_regs->iprh ));

    OSA_DPRINTF(("Switched Central Resource registers\n"));
    OSA_DPRINTF((" DAVINCI_MSTPRI0  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_MSTPRI0)));  
    OSA_DPRINTF((" DAVINCI_MSTPRI1  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_MSTPRI0)));  

    OSA_DPRINTF(("Other counters:\n"));
    OSA_DPRINTF(("  g_nCountRxPing   = %d\n", g_nCountRxPing ));
    OSA_DPRINTF(("  g_nCountRxPong   = %d\n", g_nCountRxPong ));
    OSA_DPRINTF(("  g_nCountTxPing   = %d\n", g_nCountTxPing ));
    OSA_DPRINTF(("  g_nCountTxPong   = %d\n", g_nCountTxPong ));
    
    OSA_DPRINTF(("Interrupt Controller registers\n"));
    
    OSA_DPRINTF((" DAVINCI_FIQ0     = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_FIQ0    )));  // Interrupt Status 0 [Interrupt Status of INT[31:0] (If Mapped to FIQ)]
    OSA_DPRINTF((" DAVINCI_FIQ1     = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_FIQ1    )));  // Interrupt Status 1 [Interrupt Status of INT[63:32] (If Mapped to FIQ)]
    OSA_DPRINTF((" DAVINCI_IRQ0     = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_IRQ0    )));  // Interrupt Status 0 [Interrupt Status of INT[31:0] (If Mapped to IRQ)]
    OSA_DPRINTF((" DAVINCI_IRQ1     = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_IRQ1    )));  // Interrupt Status 1 [Interrupt Status of INT[63:32] (If Mapped to IRQ)]
    OSA_DPRINTF((" DAVINCI_FIQENTRY = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_FIQENTRY)));  // Entry Address [28:0] for Valid FIQ Interrupt
    OSA_DPRINTF((" DAVINCI_IRQENTRY = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_IRQENTRY)));  // Entry Address [28:0] for Valid IRQ Interrupt
    OSA_DPRINTF((" DAVINCI_EINT0    = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_EINT0   )));  // Interrupt Enable Register 0
    OSA_DPRINTF((" DAVINCI_EINT1    = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_EINT1   )));  // Interrupt Enable Register 1
    OSA_DPRINTF((" DAVINCI_INCTL    = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_INCTL   )));  // Interrupt Operation Control Register
    OSA_DPRINTF((" DAVINCI_EABASE   = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_EABASE  )));  // Interrupt Entry Table Base Address Register
    OSA_DPRINTF((" DAVINCI_INTPRI0  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_INTPRI0 )));  // Interrupt 0-7 Priority Select
    OSA_DPRINTF((" DAVINCI_INTPRI1  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_INTPRI1 )));  // Interrupt 8-15 Priority Select
    OSA_DPRINTF((" DAVINCI_INTPRI2  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_INTPRI2 )));  // Interrupt 16-23 Priority Select
    OSA_DPRINTF((" DAVINCI_INTPRI3  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_INTPRI3 )));  // Interrupt 24-31 Priority Select. This register that holds the interrupt priority for the ASP interrupts
    OSA_DPRINTF((" DAVINCI_INTPRI4  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_INTPRI4 )));  // Interrupt 32-39 Priority Select
    OSA_DPRINTF((" DAVINCI_INTPRI5  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_INTPRI5 )));  // Interrupt 40-47 Priority Select
    OSA_DPRINTF((" DAVINCI_INTPRI6  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_INTPRI6 )));  // Interrupt 48-55 Priority Select
    OSA_DPRINTF((" DAVINCI_INTPRI7  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_INTPRI7 )));  // Interrupt 56-63 Priority Select
    
    OSA_DPRINTF(("Switched Central Resource registers\n"));
    OSA_DPRINTF((" DAVINCI_MSTPRI0  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_MSTPRI0)));  
    OSA_DPRINTF((" DAVINCI_MSTPRI1  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_MSTPRI0)));  
    
    OSA_DPRINTF(("EMIF/NAND registers\n"));
    OSA_DPRINTF((" DAVINCI_AWCCR  = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_AWCCR)));  // Asynchronous Wait Cycle Configuration Register   
    OSA_DPRINTF((" DAVINCI_A1CR   = 0x%.8X\n", OsaLinuxAspReadReg32(DAVINCI_A1CR)));    // Asynchronous 1 Configuration Register (CS2 Space)
    
    OSA_DPRINTF(("Buffer size    = %d\n", g_cbBufSize));

  } else if (nCmd == 2) {
    OsaUint8 value = 0xA5;
    OsaUint8 *pBufAddr;
    
    OSA_DPRINTF(("Filling out buffer with value=0x%.2X, \n", value));
   
    OSA_DPRINTF(("g_sqDmaBuffers[%d].virt = 0x%p \n", OSA_ASP_BUF_TX_PING, g_sqDmaBuffers[OSA_ASP_BUF_TX_PING].virt));
    OSA_DPRINTF(("g_sqDmaBuffers[%d].virt = 0x%p \n", OSA_ASP_BUF_TX_PONG, g_sqDmaBuffers[OSA_ASP_BUF_TX_PONG].virt));
    
    pBufAddr = (OsaUint8 *) g_sqDmaBuffers[OSA_ASP_BUF_TX_PING].virt;
    if(pBufAddr != NULL) {
      memset(pBufAddr, value, g_cbBufSize);
    }
    
    pBufAddr = (OsaUint8 *) g_sqDmaBuffers[OSA_ASP_BUF_TX_PONG].virt;
    if(pBufAddr != NULL) {
      memset(pBufAddr, value, g_cbBufSize);
    }

  } else if (nCmd == 3) {
    OsaUint8  *pBufAddrByte;
    OsaUint32 nBufCnt = 0;
    OsaUint32 nCnt    = 0;
    
    OSA_DPRINTF(("Fill OutBuf with a count sequence\n"));
    
    pBufAddrByte = (OsaUint8 *)  g_sqDmaBuffers[OSA_ASP_BUF_TX_PING].virt;
    if(pBufAddrByte != NULL) {
      for(nBufCnt = 0; nBufCnt < g_cbBufSize; nBufCnt++) {
        *pBufAddrByte = (OsaUint8) nCnt;
      }
      pBufAddrByte++;
      nCnt++;
    }
    
    pBufAddrByte = (OsaUint8 *)  g_sqDmaBuffers[OSA_ASP_BUF_TX_PONG].virt;
    if(pBufAddrByte != NULL) {
      for(nBufCnt = 0; nBufCnt < g_cbBufSize; nBufCnt++) {
        *pBufAddrByte = (OsaUint8) nCnt;
      }
      pBufAddrByte++;
      nCnt++;
    }

  } else if (nCmd == 4) {
    int i;

    // Dump 128 bytes from RX PING buffer
    OsaUint32 nBytes = 128;

    char *strbuf;
    dma_addr_t phys;
    
    strbuf = (char *)dma_alloc_coherent(NULL, nBytes * 6, &phys, GFP_KERNEL);
    
    if (!strbuf) {
      OSA_DPRINTF(("dma_alloc_coherent() failed \n"));
      return -1;
    }
    
    OSA_DPRINTF(("OsaLinuxAspPrintRxBuf() nBytes=%d, \n", nBytes));
    
    strbuf[0] = '\0';
    
    for (i = 0; i < nBytes; i++) {
      sprintf(strbuf + strlen(strbuf), "0x%.2X ", g_sqDmaBuffers[OSA_ASP_BUF_RX_PING].virt[i]);
    }
    OSA_DPRINTF(("%s\n", strbuf));
    
    dma_free_coherent(NULL, nBytes * 6, strbuf, phys);

  } else if (nCmd == 5) {
    OSA_DPRINTF(("Clearing pending interrupts in IPR (writing to ICR)\n"));
    if (ptr_edmacc_regs->ipr & (0x1 << OSA_ASP_TCC_RX_LINK_PONG)) {
      ptr_edmacc_regs->icr =  (0x1 << OSA_ASP_TCC_RX_LINK_PONG);
    }
    if (ptr_edmacc_regs->ipr & (0x1 << OSA_ASP_TCC_RX_LINK_PING)) {
      ptr_edmacc_regs->icr =  (0x1 << OSA_ASP_TCC_RX_LINK_PING);
    }
    if (ptr_edmacc_regs->ipr & (0x1 << OSA_ASP_TCC_TX_LINK_PONG)) {
      ptr_edmacc_regs->icr =  (0x1 << OSA_ASP_TCC_TX_LINK_PONG);
    }
    if (ptr_edmacc_regs->ipr & (0x1 << OSA_ASP_TCC_TX_LINK_PING)) {
      ptr_edmacc_regs->icr =  (0x1 << OSA_ASP_TCC_TX_LINK_PING);
    }

  } else if (nCmd == 6) {
    OSA_DPRINTF(("Setting EER bit 0x80000000 (writing to EESR)\n"));
    ptr_edmacc_regs->eesr =  (0x1 << 31);

  } else if (nCmd == 7) {
    int i;

    OSA_DPRINTF(("Complete edmacc_paramentry_regs dump:\n"));
    for (i = 0; i < 128; i++) {
      OSA_DPRINTF(("[%d] address = 0x%p \n", i, &(ptr_edmacc_regs->paramentry[i])));
      OSA_DPRINTF(("  opt          = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].opt ));
      OSA_DPRINTF(("  src          = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].src ));
      OSA_DPRINTF(("  a_b_cnt      = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].a_b_cnt ));
      OSA_DPRINTF(("  dst          = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].dst ));
      OSA_DPRINTF(("  src_dst_bidx = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].src_dst_bidx ));
      OSA_DPRINTF(("  link_bcntrld = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].link_bcntrld ));
      OSA_DPRINTF(("  src_dst_cidx = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].src_dst_cidx ));
      OSA_DPRINTF(("  ccnt         = 0x%.8X\n", ptr_edmacc_regs->paramentry[i].ccnt ));
    }


  } else {
    OSA_DPRINTF(("Unsupported Dbg Cmd (%d)\n", nCmd));
  }

  return 0;
  
}


/******************************************************************************
  OsaLinuxDmpReadReg32
    return 0=>OK, -1=>error
******************************************************************************/
static inline u32 OsaLinuxAspReadReg32(u32 address)
{
  return *((volatile u32*) io_p2v(address));
}

/******************************************************************************
  OsaLinuxDmpReadReg32
    return 0=>OK, -1=>error
******************************************************************************/
static inline void OsaLinuxAspWriteReg32(u32 address, u32 value)
{
  *((volatile u32*) io_p2v(address)) = value;
}


#endif // DAVINCI
